import logging
import re
import sys
import yaml

from io import IOBase

from truffleHog3.types import Any, File, List, Regexes


__all__ = ("logger", "load", "dump", "compile", "match")


_LOGGER_NAME = "trufflehog"
_LOGGER_FMT = "[%(levelname).1s] %(message)s"


def logger(level: int = logging.ERROR) -> logging.Logger:
    logger = logging.Logger.manager.loggerDict.get(_LOGGER_NAME)
    if logger:
        logger.setLevel(level)
        return logger

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(logging.Formatter(_LOGGER_FMT))
    logger = logging.getLogger(_LOGGER_NAME)
    logger.setLevel(level)
    logger.addHandler(stream_handler)
    return logger


def load(file: File) -> Any:
    if not isinstance(file, IOBase):
        file = open(file)
    return yaml.safe_load(file)


def dump(obj: Any, file: File = None):
    if file is None:
        file = sys.stdout
    elif not isinstance(file, IOBase):
        file = open(file, "w")
    file.write(str(obj))


def compile(raw: List[str]) -> Regexes:
    return [re.compile(s) for s in raw]


def match(string: str, regexes: Regexes) -> str:
    for regex in regexes:
        if regex.search(string):
            return regex.pattern
    return None



import git
import os

from abc import ABC, abstractmethod
from datetime import datetime

from truffleHog3.lib import log, utils
from truffleHog3.types import List, MetaGen, Regexes


_FMT = "%Y-%m-%d %H:%M:%S"
_NOW = datetime.now().strftime(_FMT)

SELF = ["trufflehog.json", "trufflehog.yaml", "trufflehog.yml"]


class Engine(ABC):
    def __init__(self, path: str, skip: List[str] = None):
        self.path = path
        self.regexes = skip

    @property
    def regexes(self) -> Regexes:
        return self._regexes

    @regexes.setter
    def regexes(self, skip: List[str]):
        self._regexes = utils.compile((skip or []) + SELF)

    @abstractmethod
    def __iter__(self) -> MetaGen:
        ...  # pragma: no cover

    def skip(self, path: str) -> bool:
        regex = utils.match(path, self.regexes)
        if regex:
            log.info(f"skipping '{path}' matched by '{regex}'")
            return True

        return False


class Simple(Engine):
    def __iter__(self) -> MetaGen:
        for root, dirs, files in os.walk(self.path):
            for d in dirs:
                if d.startswith(".") or self.skip(d):
                    dirs.remove(d)

            for file in files:
                filepath = os.path.join(root, file)
                relative_path = os.path.relpath(filepath, self.path)
                if self.skip(relative_path):
                    continue

                try:
                    with open(filepath) as f:
                        data = f.read()
                except Exception as error:  # pragma: no cover
                    log.warning(f"error reading '{filepath}': {error}")
                    continue

                yield {
                    "date": _NOW,
                    "path": relative_path,
                    "branch": None,
                    "commit": None,
                    "commitHash": None,
                    "data": data,
                }


class Git(Engine):
    def __init__(
        self,
        path: str,
        branch: str = None,
        since_commit: str = None,
        max_depth: int = None,
        **kwargs,
    ):
        self.repo = git.Repo(path)
        self.branch = branch
        self.since_commit = since_commit
        self.max_depth = max_depth
        super().__init__(path, **kwargs)

    def __iter__(self):
        already_searched = set()

        for branch in self._get_branches():
            log.info(f"switching to branch '{branch}'")
            prev_commit = None
            since_commit_reached = False
            commits = self.repo.iter_commits(branch, max_count=self.max_depth)

            for curr_commit in commits:
                if curr_commit.hexsha == self.since_commit:
                    since_commit_reached = True

                if self.since_commit and since_commit_reached:
                    prev_commit = curr_commit
                    continue

                diff_hash = str(prev_commit) + str(curr_commit)
                if not prev_commit or diff_hash in already_searched:
                    prev_commit = curr_commit
                    continue

                diff = prev_commit.diff(curr_commit, create_patch=True)
                already_searched.add(diff_hash)
                yield from self._diff_worker(diff, prev_commit, branch)
                prev_commit = curr_commit

            diff = curr_commit.diff(git.NULL_TREE, create_patch=True)
            yield from self._diff_worker(diff, prev_commit, branch)

    def _diff_worker(
        self, diff: git.DiffIndex, commit: git.Commit, branch: git.Head = None
    ) -> MetaGen:
        for blob in diff:
            pdiff = blob.diff.decode("utf-8", errors="replace")
            path = blob.b_path if blob.b_path else blob.a_path

            if pdiff.startswith("Binary files") or self.skip(path):
                continue

            date = datetime.fromtimestamp(commit.committed_date)
            yield {
                "date": date.strftime(_FMT),
                "path": path,
                "branch": branch.name,
                "commit": commit.message,
                "commitHash": commit.hexsha,
                "data": pdiff,
            }

    def _get_branches(self) -> List[git.Head]:
        if self.repo.remotes:  # pragma: no cover
            try:
                return self.repo.remotes.origin.fetch(self.branch)
            except git.GitCommandError as error:
                log.warning(f"error fetching remote branches: {error}")
        else:
            log.warning("missing remote")

        if not self.branch:
            branches = [self.repo.active_branch]
        else:
            branches = [self.repo.branches[self.branch]]

        return branches



import yaml

from typing import Any, Generator, List, Mapping, re, TextIO, Tuple, Union


File = Union[str, TextIO]
Issue = Mapping[str, Any]
Issues = List[Issue]
Meta = Mapping[str, Any]
MetaGen = Generator[Meta, None, None]
Regex = re.Pattern
Regexes = List[re.Pattern]
Repo = Tuple[str, Meta]
RawRules = Mapping[str, str]
Rules = Mapping[str, re.Pattern]
SkipRules = Union[List[str], Mapping[str, List[str]]]
StrGen = Generator[str, None, None]


class Config:
    def __init__(self, **kwargs):
        self.branch = None
        self.since_commit = None
        self.skip_strings = []
        self.skip_paths = []
        self.line_numbers = False
        self.max_depth = 1000000
        self.no_regex = False
        self.no_entropy = False
        self.no_history = False
        self.no_current = False
        self.update(**kwargs)

    def __repr__(self):
        return yaml.safe_dump(self.raw, default_flow_style=False)

    def __eq__(self, other):
        if isinstance(other, Config):
            return self.raw == other.raw
        return self.raw == other

    @property
    def raw(self) -> dict:
        items = self.__dict__.items()
        return {k: v for k, v in items if not k.startswith("_")}

    def update(self, **kwargs):
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)



import jinja2
import json
import os
import yaml

from abc import ABC, abstractmethod
from collections import defaultdict

from truffleHog3.types import Issue, Issues, Meta


__all__ = ("TEXT", "JSON", "YAML", "HTML")


class _colors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


_TEXT_TEMPLATE = """~~~~~~~~~~~~~~~~~~~~~%s
Reason: {reason}
Path:   {path}
Branch: {branch}
Commit: {commit}
Hash:   {commitHash}
%s{strings}
~~~~~~~~~~~~~~~~~~~~~"""

_HTML_STATIC_FOLDER = os.path.dirname(os.path.dirname(__file__))
_HTML_TEMPLATE = "report.j2"


class Engine(ABC):
    def __init__(self, issues: Issues):
        self.issues = issues

    @abstractmethod
    def __str__(self):
        ...  # pragma: no cover


class TEXT(Engine):
    def __init__(self, issues: Issues):
        self.load_template()
        super().__init__(issues)

    def __str__(self):
        return "\n".join(self._render(i) for i in self.issues)

    def _render(self, issue: Issue) -> str:
        strings = "\n".join(issue["stringsFound"])
        return self.template.format(**issue, strings=strings)

    def load_template(self):
        self.template = _TEXT_TEMPLATE % (_colors.OKGREEN, _colors.ENDC)


class JSON(Engine):
    def __str__(self):
        return json.dumps(self.issues, indent=2)


class YAML(Engine):
    def __str__(self):
        return yaml.safe_dump(self.issues, default_flow_style=False)


class HTML(Engine):
    def __init__(self, issues: Issues):
        self.load_template()
        super().__init__(issues)

    def __str__(self):
        return self.template.render(report=_make_report(self.issues))

    def load_template(self):
        env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(_HTML_STATIC_FOLDER),
            autoescape=True,
        )
        env.filters["strip"] = _strip
        self.template = env.get_template(_HTML_TEMPLATE)


def _make_report(issues: Issues) -> Meta:
    total, entropy, regexes = len(issues), 0, 0
    mapping = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))

    for issue in issues:
        reason = issue["reason"]
        if reason == "High entropy":
            entropy += 1
        else:
            regexes += 1

        folder = os.path.dirname(issue["path"])
        filename = os.path.basename(issue["path"])
        mapping[reason][folder][filename].append(issue)

    return {
        "total": total,
        "entropy": entropy,
        "regexes": regexes,
        "issues": mapping,
    }


def _strip(data: str) -> str:
    return data.strip() if data else None
